/*$T indentinput.cpp GC 1.140 04/13/11 21:23:07 */

/*$6
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

 Modify and cleanup by Binh Nguyen.

 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 */

#include <sys/time.h>
#include <iostream>
#include <vector>
#include <bitset>
#include <ctime>
#include <stdlib.h>
#include <fstream>
#include "hash_element.h"

using namespace std;

/*
 =======================================================================================================================
 =======================================================================================================================
 */

void print_list(vector<tIpAddress > &some_list)
{
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	vector<tIpAddress >::iterator	it;
	tIpAddress::iterator			it_inner;
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	for(it = some_list.begin(); it != some_list.end(); it++)
	{
		for(it_inner = it->begin(); it_inner != it->end(); it_inner++)
		{
			cout << *it_inner << " ";
		}

		cout << endl;
	}
}

/*
 =======================================================================================================================
 =======================================================================================================================
 */
int hash_function(tIpAddress &ip, int N)
{
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	tIpAddress::iterator	it_inner;
	int										hash_value = 0;
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	for(it_inner = ip.begin(); it_inner != ip.end(); it_inner++)
	{
		hash_value += *it_inner;
	}

	return(hash_value % N);
}

/*
 =======================================================================================================================
 =======================================================================================================================
 */
int main(int argc, char *argv[])
{
	if(argc != 7)
	{
		cerr <<
			"usage: ./chain_hash <ip_file> <ip_test> <table_size> <num_elements> <num_test_ele> <0 not to print, other # implies print>" <<
			endl;
		exit(1);
	}

	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	/* Parameters */
	int num_ips = 0;
	int num_elements = atoi(argv[4]);
	int num_ip_components;
	int num_test_ele = atoi(argv[5]);
	int print_choice = atoi(argv[6]);
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	num_ip_components = 4;

	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	/* Table size */
	int												N = atoi(argv[3]);
	/* General variables */
	int												i, j, hash_value;
	vector<tIpAddress >::iterator	it, it2;
	tIpAddress::iterator			it_inner, it_inner2;
	/* Populate the IPs */
	vector<tIpAddress >				ip_list, ip_test_list;
	ifstream										fin, tfin;
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	fin.open(argv[1]);
	if(!fin)
	{
		cerr << "Error: " << argv[1] << " could not be opened" << endl;
		exit(1);
	}

	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	string::size_type	gap_pos, sub_length;
	string				line, sub_ip_str;
	int					num_test_ip = -1;
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	num_ips = -1;
	while(!fin.eof() && num_ips < num_elements)
	{
		getline(fin, line);

		/*~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
		tIpAddress	ip;
		/*~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

		i = 0;
		for(j = 0; j < num_ip_components; j++)
		{
			gap_pos = line.find_first_of(" ", i);
			sub_length = gap_pos - i;
			sub_ip_str = line.substr(i, sub_length);
			ip.push_back(atoi(sub_ip_str.c_str()));
			i = gap_pos + 1;
		}

		ip_list.push_back(ip);
		num_ips++;
	}

	fin.close();
	tfin.open(argv[2]);
	if(!tfin)
	{
		cerr << "Error: " << argv[2] << " could not be opened" << endl;
		exit(1);
	}

	while(!tfin.eof() && num_test_ip < num_test_ele)
	{
		getline(tfin, line);

		/*~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
		tIpAddress	ip;
		/*~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

		i = 0;
		for(j = 0; j < num_ip_components; j++)
		{
			gap_pos = line.find_first_of(" ", i);
			sub_length = gap_pos - i;
			sub_ip_str = line.substr(i, sub_length);
			ip.push_back(atoi(sub_ip_str.c_str()));
			i = gap_pos + 1;
		}

		ip_test_list.push_back(ip);
		num_test_ip++;
	}

	tfin.close();
	if(print_choice > 0)
	{
		cout << "Printing IPs..." << endl;
		print_list(ip_list);
	}

	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	/* Put IPs on Hash */
	Hash_element*	hashTable = new Hash_element[N];
	/* get the time right before we begin inserting elements DO NOT CHANGE */
	struct timeval	pre_insert, post_insert, post_algo;
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	gettimeofday(&pre_insert, NULL);

	// inserting ip list into hash table
	for(it = ip_list.begin(); it != ip_list.end(); it++)
	{
		hash_value = hash_function(*it, N); // hashing
		hashTable[hash_value].insertValue(*it);
	}

	/* get the time right after we finish inserting elements DO NOT CHANGE */
	gettimeofday(&post_insert, NULL);

	/*~~~~~~~~~~~~~~*/
	int hit_count = 0;
	/*~~~~~~~~~~~~~~*/

	for(it = ip_test_list.begin(); it != ip_test_list.end(); it++)
	{
		hash_value = hash_function(*it, N);
		if(hashTable[hash_value].isValueContained(*it))
		{
			hit_count++;
		}
	}

	/* Finish up and return. DO NOT CHANGE */
	gettimeofday(&post_algo, NULL);

	delete [] hashTable;

	/*~~~~~~~~~~~~~~~~~~~~~~*/
	long	seconds, useconds;
	/*~~~~~~~~~~~~~~~~~~~~~~*/

	seconds = post_insert.tv_sec - pre_insert.tv_sec;
	useconds = post_insert.tv_usec - pre_insert.tv_usec;

	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	double	insert_time = ((seconds) * 1000 + useconds / 1000.0) + 0.5;
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	seconds = post_algo.tv_sec - post_insert.tv_sec;
	useconds = post_algo.tv_usec - post_insert.tv_usec;

	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	double	test_time = ((seconds) * 1000 + useconds / 1000.0) + 0.5;
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	cout << "insert_time=" << insert_time << " test_time = "
		<< test_time << " num false positive " << hit_count << endl;

	if(print_choice > 0)
	{
		cout << "Printing filter table..." << endl;

		/*~~~~~~~~~~~*/
		int hash_index;
		/*~~~~~~~~~~~*/

		for(hash_index = 0; hash_index < N; hash_index++)
		{
			if(hashTable[hash_index].getSize() > 0)
			{
				cout << "-------------------------" << endl;
				hashTable[hash_index].print();
			}
		}
	}

	return 0;
}
